#include "../core/utils/utils.h"
#include "../core/debug/evemon.h"
#include "../core/debug/exception.h"
#include "../core/math/solve/base.h"
#include "../core/math/solve/fixed.h"
#include "../core/math/solve/math.h"
#include "../core/math/matrix.h"
#include "../core/math/symbolic/var.h"
#include "../core/math/symbolic/matrix.h"
#include "../core/math/symbolic/collection.h"
#include "../core/math/optics/paraxial/abcd.h"
#include "../core/math/optics/paraxial/yu.h"
#include "../core/math/optics/paraxial/seidel.h"
#include "../core/math/optics/paraxial/wavefront.h"

#include "refractive.h"
#include "thickness.h"
#include "thicklens.h"

#include <cstdio>
#include <cmath>
#include <array>
#include <algorithm>

// #define BENCHMARK

// compute total sensitivity
double get_sensitivity(const math::symbolic::collection<double>& collection, const math::symbolic::var<double>& var)
{
	double sumsq = 0;

	// retrieve all names (recursive)
	auto names = collection.list_all_names();

	// browse list
	for (auto& v : names)
	{
		// get current sensitivity
		auto curr = collection.get_deviation(var, v);

		// add to cumulative sum
		sumsq += curr * curr;
	}

	// return rms
	return sqrt(sumsq);
}

// display sensitivity
void display_sensitivity_tree(const double total_variance, const math::symbolic::collection<double>& collection, const std::string& name, const math::symbolic::var<double>& var, const std::string& units, const double scale, int iter=0)
{
	// compute local variance
	{
		auto local_var = get_sensitivity(collection, var);

		// display collection variance
#ifndef BENCHMARK
		for (int i = 0; i < iter; i++)
			printf("   ");

		printf("+ d%s/d%s\t%.5le %s\t%.1f%%\r\n", name.c_str(), collection.name().c_str(), local_var * scale, units.c_str(), 1e2 * std::pow(local_var / total_variance, 2));
#endif
	}

	// display local items
	{
		auto list = collection.list_local_names();

		for (auto& v : list)
		{
			// compute current variance
			auto curr_var = collection.get_deviation(var, v);

			// display
#ifndef BENCHMARK
			for (int i = 0; i <= iter; i++)
				printf("   ");

			printf("- d%s/d%s\t%.5le %s\t%.1f%%\r\n", name.c_str(), v.c_str(), curr_var * scale, units.c_str(), 1e2 * std::pow(curr_var / total_variance, 2));
#endif
		}
	}

	// display children
	{
		auto list = collection.list_subgroups();

		for (auto& v : list)
			display_sensitivity_tree(total_variance, collection.get_subgroup(v), name, var, units, scale, iter + 1);
	}

}

// display sensitivity
void display_sensitivity(const math::symbolic::collection<double>& collection, const std::string& name, const math::symbolic::var<double>& var, const std::string& units, const double scale=1)
{
	auto eval_value = var.eval();
	auto total_variance = get_sensitivity(collection, var);

	// print value
#ifndef BENCHMARK
	printf("%s = %f %s\r\n", name.c_str(), eval_value * scale, units.c_str());
	printf("d%s = %.5le %s (%.1f%%)\r\n", name.c_str(), total_variance * scale, units.c_str(), 1e2 * total_variance / eval_value);
#endif

	// display tree
	display_sensitivity_tree(total_variance, collection, name, var, units, scale);

	// space a bit the display
#ifndef BENCHMARK
	printf("\r\n");
#endif
}

// create sample system
auto create_sample_system(math::symbolic::collection<double> &collection)
{
	// consts
	auto air = math::symbolic::var<double>(1);
	auto zero = math::symbolic::var<double>();

	// lens 1
	auto& lens1_group = collection.create_subgroup("lens1");

	auto lens1_r1 = lens1_group.add("lens1:R1", math::symbolic::var<double>(-5.091000000000000E+001), 1e-2, math::symbolic::sensitivity::proportional);
	auto lens1_r2 = lens1_group.add("lens1:R2", math::symbolic::var<double>(2.840600000000000E+001), 1e-2, math::symbolic::sensitivity::proportional);
	auto lens1_th = lens1_group.add("lens1:TH", math::symbolic::var<double>(1.000000000000000E+000), 50e-6, math::symbolic::sensitivity::absolute);
	auto lens1_ni = lens1_group.add("lens1:ni", math::symbolic::var<double>(1.5187), 2e-4, math::symbolic::sensitivity::absolute);

	// lens 2
	auto& lens2_group = collection.create_subgroup("lens2");

	auto lens2_r1 = lens2_group.add("lens2:R1", math::symbolic::var<double>(-4.626200000000000E+001), 1e-2, math::symbolic::sensitivity::proportional);
	auto lens2_r2 = lens2_group.add("lens2:R2", math::symbolic::var<double>(-3.017800000000000E+001), 1e-2, math::symbolic::sensitivity::proportional);
	auto lens2_th = lens2_group.add("lens2:TH", math::symbolic::var<double>(2.631000000000000E+000), 50e-6, math::symbolic::sensitivity::absolute);
	auto lens2_ni = lens2_group.add("lens2:ni", math::symbolic::var<double>(1.5187), 2e-4, math::symbolic::sensitivity::absolute);

	// lens 3
	auto& lens3_group = collection.create_subgroup("lens3");

	auto lens3_r1 = lens3_group.add("lens3:R1", math::symbolic::var<double>(1.562490000000000E+002), 1e-2, math::symbolic::sensitivity::proportional);
	auto lens3_r2 = lens3_group.add("lens3:R2", math::symbolic::var<double>(-1.725100000000000E+002), 1e-2, math::symbolic::sensitivity::proportional);
	auto lens3_th = lens3_group.add("lens3:TH", math::symbolic::var<double>(2.548000000000000E+000), 50e-6, math::symbolic::sensitivity::absolute);
	auto lens3_ni = lens3_group.add("lens3:ni", math::symbolic::var<double>(1.5187), 2e-4, math::symbolic::sensitivity::absolute);

	// lens 4
	auto& lens4_group = collection.create_subgroup("lens4");

	auto lens4_r1 = lens4_group.add("lens4:R1", math::symbolic::var<double>(-6.013150000000001E+002), 1e-2, math::symbolic::sensitivity::proportional);
	auto lens4_r2 = lens4_group.add("lens4:R2", math::symbolic::var<double>(-1.237470000000000E+002), 1e-2, math::symbolic::sensitivity::proportional);
	auto lens4_th = lens4_group.add("lens4:TH", math::symbolic::var<double>(1.819000000000000E+000), 50e-6, math::symbolic::sensitivity::absolute);
	auto lens4_ni = lens4_group.add("lens4:ni", math::symbolic::var<double>(1.5187), 2e-4, math::symbolic::sensitivity::absolute);

	// thicknesses
	auto L1 = collection.add("L1", math::symbolic::var<double>(3.706400000000000E+001), 50e-6, math::symbolic::sensitivity::absolute);
	auto L2 = collection.add("L2", math::symbolic::var<double>(4.348600000000000E+001), 50e-6, math::symbolic::sensitivity::absolute);
	auto L3 = collection.add("L3", math::symbolic::var<double>(3.000000000000000E+000), 50e-6, math::symbolic::sensitivity::absolute);
	auto L4 = collection.add("L4", math::symbolic::var<double>(1.986000000000000E+000), 50e-6, math::symbolic::sensitivity::absolute);

	// build system
	math::optics::system::group<double> system;

	auto lens1 = std::make_shared<thicklens<double>>(lens1_r1, lens1_r2, lens1_th, lens1_ni, air);
	auto lens2 = std::make_shared<thicklens<double>>(lens2_r1, lens2_r2, lens2_th, lens2_ni, air);
	auto lens3 = std::make_shared<thicklens<double>>(lens3_r1, lens3_r2, lens3_th, lens3_ni, air);
	auto lens4 = std::make_shared<thicklens<double>>(lens4_r1, lens4_r2, lens4_th, lens4_ni, air);

	// insert initial spacing (telecentric lens)
	system.push_back(std::make_shared<thickness_surface<double>>(L1));

	// insert lens 1
	system.push_back(std::make_shared<thickness_surface<double>>(math::symbolic::var<double>(lens1->get_paraxial_matrix().ffl() - lens1->get_paraxial_matrix().efl())));
	system.push_back(lens1);
	system.push_back(std::make_shared<thickness_surface<double>>(math::symbolic::var<double>(lens1->get_paraxial_matrix().bfl() - lens1->get_paraxial_matrix().efl())));

	// insert spacing L2
	system.push_back(std::make_shared<thickness_surface<double>>(L2));

	// insert lens 2
	system.push_back(std::make_shared<thickness_surface<double>>(math::symbolic::var<double>(lens2->get_paraxial_matrix().ffl() - lens2->get_paraxial_matrix().efl())));
	system.push_back(lens2);
	system.push_back(std::make_shared<thickness_surface<double>>(math::symbolic::var<double>(lens2->get_paraxial_matrix().bfl() - lens2->get_paraxial_matrix().efl())));

	// insert spacing L3
	system.push_back(std::make_shared<thickness_surface<double>>(L3));

	// insert lens 3
	system.push_back(std::make_shared<thickness_surface<double>>(math::symbolic::var<double>(lens3->get_paraxial_matrix().ffl() - lens3->get_paraxial_matrix().efl())));
	system.push_back(lens3);
	system.push_back(std::make_shared<thickness_surface<double>>(math::symbolic::var<double>(lens3->get_paraxial_matrix().bfl() - lens3->get_paraxial_matrix().efl())));

	// insert spacing L4
	system.push_back(std::make_shared<thickness_surface<double>>(L4));

	// insert lens 4
	system.push_back(std::make_shared<thickness_surface<double>>(math::symbolic::var<double>(lens4->get_paraxial_matrix().ffl() - lens4->get_paraxial_matrix().efl())));
	system.push_back(lens4);
	system.push_back(std::make_shared<thickness_surface<double>>(math::symbolic::var<double>(lens4->get_paraxial_matrix().bfl() - lens4->get_paraxial_matrix().efl())));

	return system;
}

// main program
int main(int argc, char** argv)
{
	try
	{
		
#ifdef BENCHMARK
		const size_t N = 100;
#else
		const size_t N = 1;
#endif

		std::array<double, N> results;

		for (size_t iter = 0; iter < N; iter++)
		{
			// get initial time
			auto tick1 = std::chrono::high_resolution_clock::now();

			// create sample system
			math::symbolic::collection<double> collection("all");

			auto system = create_sample_system(collection);

			// display BFL
			{
				auto M = system.get_paraxial_matrix();

				auto BFL = M.bfl();

				display_sensitivity(collection, "BFL", BFL, "mm");
			}

			// compute seidel aberrations
			math::optics::paraxial::yu<double> yu_chief(math::symbolic::var<double>(0), math::symbolic::var<double>(0.01745329251994329576923690768489));
			math::optics::paraxial::yu<double> yu_marginal(math::symbolic::var<double>(5.0), math::symbolic::var<double>(0));

			auto seidel_tot = system.get_seidel_coefficients(yu_marginal, yu_chief);

			// display sensitivity
			{
				display_sensitivity(collection, "spherical", seidel_tot.spherical(), "um", 1e3);
				display_sensitivity(collection, "coma", seidel_tot.coma(), "um", 1e3);
				display_sensitivity(collection, "astig", seidel_tot.astig(), "um", 1e3);
				display_sensitivity(collection, "petzval", seidel_tot.petzval(), "um", 1e3);
				display_sensitivity(collection, "dist", seidel_tot.dist(), "um", 1e3);
			}

			// convert to wavefront
			auto wfe = seidel_tot.to_wavefront();

			for (double y : {0.0, 1.0})
			{
				auto rmsval = wfe.rms(y);

				// display value
				display_sensitivity(collection, format("RMS(%.3f)", y), rmsval, "lambdas", 1.0 / 532e-6);
			}

			// get current time
			auto tick2 = std::chrono::high_resolution_clock::now();

			// compute time
			auto delta_ticks = std::chrono::duration_cast<std::chrono::duration<double>>(tick2 - tick1);

			auto value = 1e3 * delta_ticks.count();

			results[iter] = value;

			printf("%.3f ms\r\n", value);
		}

		printf("\r\n");

		// compute mean & display
		std::sort(results.begin(), results.end());

		double median;
		
		if((N % 2) == 1)
			median = results[N >> 1];
		else
			median = (results[(N >> 1) - 1] + results[N >> 1]) * 0.5;

		printf("Probable time: %.3f ms\r\n", median);

		// return successfuly
		return 0;
	}
	catch (exception::base& e)
	{
		printf("error: %s\r\n", e.to_string().c_str());

		_critical("%s", e.to_string().c_str());
	}
	catch (std::exception& e)
	{
		printf("error: %s\r\n", e.what());

		_critical("%s", e.what());
	}
	catch (...)
	{
		printf("unknown error!\r\n");

		_critical("unknown error!");
	}

	// signal error
	return 1;
}
